"use strict";
var __defProp = Object.defineProperty;
var __getOwnPropDesc = Object.getOwnPropertyDescriptor;
var __getOwnPropNames = Object.getOwnPropertyNames;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __export = (target, all) => {
  for (var name in all)
    __defProp(target, name, { get: all[name], enumerable: true });
};
var __copyProps = (to, from, except, desc) => {
  if (from && typeof from === "object" || typeof from === "function") {
    for (let key of __getOwnPropNames(from))
      if (!__hasOwnProp.call(to, key) && key !== except)
        __defProp(to, key, { get: () => from[key], enumerable: !(desc = __getOwnPropDesc(from, key)) || desc.enumerable });
  }
  return to;
};
var __toCommonJS = (mod) => __copyProps(__defProp({}, "__esModule", { value: true }), mod);

// index.ts
var rushlight_server_exports = {};
__export(rushlight_server_exports, {
  CollabServer: () => CollabServer
});
module.exports = __toCommonJS(rushlight_server_exports);
var import_state = require("@codemirror/state");
var import_client2 = require("@redis/client");

// models.ts
var import_client = require("@redis/client");
var addUpdates = (0, import_client.defineScript)({
  NUMBER_OF_KEYS: 1,
  SCRIPT: `
    local version = ARGV[1]
    for i = 2, #ARGV do
      redis.call("XADD", KEYS[1], (version + i - 1) .. "-0", "d", ARGV[i])
    end
  `,
  transformArguments: () => []
  // not needed
});
var trimMayDelete = (0, import_client.defineScript)({
  NUMBER_OF_KEYS: 1,
  SCRIPT: `
    local version = ARGV[1]
    redis.call("XTRIM", KEYS[1], "MINID", "~", (version + 1) .. "-0", "LIMIT", "0")
    if redis.call("XLEN", KEYS[1]) == 0 then
      redis.call("DEL", KEYS[1])
      return true
    end
  `,
  transformArguments: () => []
  // not needed
});
var Streams = class {
  constructor(redis, prefix) {
    this.redis = redis;
    this.prefix = prefix;
  }
  /** Read document updates newer than a version, optionally blocking if there are none. */
  async read(id, version, blocking) {
    const entries = await this.redis.executeIsolated(
      (redis) => redis.xRead(
        { key: `${this.prefix}:${id}`, id: `${version}-0` },
        { BLOCK: blocking, COUNT: 1024 }
      )
    );
    const updates = entries?.[0]?.messages ?? [];
    return updates.map((u) => ({
      version: parseInt(u.id.slice(0, -2)),
      update: JSON.parse(u.message.d)
    }));
  }
  /** Add document updates at a version, returning whether it was successful. */
  async add(id, version, updates) {
    try {
      await this.redis.executeScript(addUpdates, [
        `${this.prefix}:${id}`,
        version.toString(),
        ...updates.map((u) => JSON.stringify(u))
      ]);
      return true;
    } catch (err) {
      if (err instanceof import_client.ErrorReply) {
        if (err.message.startsWith("ERR ")) {
          return false;
        }
      }
      throw err;
    }
  }
  /** Remove document updates older than a version. */
  async trim(id, version) {
    const resp = await this.redis.executeScript(trimMayDelete, [
      `${this.prefix}:${id}`,
      version.toString()
    ]);
    return Boolean(resp);
  }
};
var Dirty = class {
  constructor(redis, prefix) {
    this.redis = redis;
    this.prefix = prefix;
  }
  /** Mark a document for compaction at a given timestamp. */
  async enqueue(id, time, notExists) {
    await this.redis.zAdd(
      `${this.prefix}-dirty`,
      { score: time, value: id },
      notExists ? { NX: true } : void 0
    );
  }
  /** Return the next document that is marked for compaction. */
  async dequeue() {
    const resp = await this.redis.zPopMin(`${this.prefix}-dirty`);
    return resp ? { id: resp.value, time: resp.score } : null;
  }
};

// index.ts
var defualtOptions = {
  compactionInterval: 1e3 * 30,
  // 30 seconds
  blockingMs: 1e3 * 5,
  // 5 seconds
  blockingPoolSize: 2048,
  keyPrefix: "doc"
};
var CollabServer = class _CollabServer {
  constructor(options, streams, dirty) {
    this.options = options;
    this.streams = streams;
    this.dirty = dirty;
  }
  /** Construct a new server. This will connect to Redis. */
  static async of(options) {
    const opts = { ...defualtOptions, ...options };
    const redis = (0, import_client2.createClient)({
      url: opts.redisUrl,
      isolationPoolOptions: {
        max: opts.blockingPoolSize ?? 2048
      }
    });
    redis.on("error", console.error);
    await redis.connect();
    return new _CollabServer(
      opts,
      new Streams(redis, opts.keyPrefix),
      new Dirty(redis, opts.keyPrefix)
    );
  }
  /** Handle an incoming collaboration message. */
  handle(id, message) {
    if (message.type == "pullUpdates") {
      return this.pullUpdates(id, message.version);
    } else if (message.type == "pushUpdates") {
      return this.pushUpdates(id, message.version, message.updates);
    } else if (message.type == "getDocument") {
      return this.getDocument(id);
    } else {
      throw new Error("Invalid message type");
    }
  }
  /** Read the latest version of a document from the server. */
  async getDocument(id) {
    const { version, doc } = await this.getDocumentInternal(id);
    return { version, doc };
  }
  /** Periodically checkpoint old documents by reading the dirty list. */
  async compactionTask() {
    while (true) {
      try {
        await this.runCompaction();
      } catch (err) {
        console.error("Error during compaction:", err.toString());
      }
      await new Promise(
        (resolve) => setTimeout(resolve, 0.1 * this.options.compactionInterval)
      );
    }
  }
  async getDocumentInternal(id) {
    let { version, doc: docS } = await this.options.loadCheckpoint(id);
    let doc = import_state.Text.of(docS.split("\n"));
    const updates = await this.streams.read(id, version);
    if (updates.length && updates[0].version !== version + 1) {
      throw new Error("Invariant violated, document is desynchronized");
    }
    for (const u of updates) {
      try {
        doc = import_state.ChangeSet.fromJSON(u.update.changes).apply(doc);
      } catch (err) {
        console.error("Error applying update:", err.toString());
      }
    }
    return {
      dbVersion: version,
      version: version + updates.length,
      doc: doc.toString()
    };
  }
  async pullUpdates(id, version) {
    const updates = await this.streams.read(id, version, 5e3);
    if (updates.length && updates[0].version !== version + 1) {
      return { status: "desync" };
    }
    return { status: "ok", updates: updates.map((u) => u.update) };
  }
  async pushUpdates(id, version, updates) {
    if (!await this.streams.add(id, version, updates)) {
      return false;
    }
    const time = Date.now() + this.options.compactionInterval;
    await this.dirty.enqueue(id, time, true);
    return true;
  }
  async runCompaction() {
    while (true) {
      const resp = await this.dirty.dequeue();
      if (resp === null)
        break;
      const { id, time } = resp;
      if (Date.now() < time) {
        await this.dirty.enqueue(id, time);
        break;
      }
      try {
        const { dbVersion, version, doc } = await this.getDocumentInternal(id);
        if (dbVersion < version) {
          await this.options.saveCheckpoint(id, { version, doc });
        }
        if (!await this.streams.trim(id, dbVersion)) {
          const time2 = Date.now() + this.options.compactionInterval;
          await this.dirty.enqueue(id, time2);
        }
      } catch (err) {
        await this.dirty.enqueue(id, time);
        throw err;
      }
    }
  }
};
// Annotate the CommonJS export names for ESM import in node:
0 && (module.exports = {
  CollabServer
});
