type Checkpoint = {
    version: number;
    doc: string;
};
type Message = {
    type: "pullUpdates";
    version: number;
} | {
    type: "pushUpdates";
    version: number;
    updates: any[];
} | {
    type: "getDocument";
};
type ServerOptions = {
    /** Connection string for Redis, used to share document changes. */
    readonly redisUrl: string;
    /** Load the most recent document checkpoint from a persistent database. */
    readonly loadCheckpoint: (id: string) => Promise<Checkpoint>;
    /** Save a document checkpoint to a persistent database. */
    readonly saveCheckpoint: (id: string, checkpoint: Checkpoint) => Promise<void>;
    /**
     * How often to compact the document history, in milliseconds.
     *
     * This is the interval at which streams are checkpointed, regardless of how
     * many server replicas are running. Each modified document is queued for
     * compaction. On each interval, the database is updated to reflect new
     * changes, while the updates older than the previous version are removed from
     * the document's Redis stream.
     *
     * The tradeoff here is between memory usage and performance. A smaller
     * interval will result in more frequent compaction, lowering Redis's memory
     * usage, but requiring more frequent database updates and potentially
     * desynchronizing with slower clients. A larger interval will mean updates
     * are stored for longer, putting more memory pressure on Redis.
     *
     * The default value balances these considerations.
     */
    readonly compactionInterval?: number;
    /** How long to block on a stream when long polling for updates. */
    readonly blockingMs?: number;
    /** The maximum number of Redis connections for long polling. */
    readonly blockingPoolSize?: number;
    /** Prefix used for keys in Redis. */
    readonly keyPrefix?: string;
};
declare const defualtOptions: {
    compactionInterval: number;
    blockingMs: number;
    blockingPoolSize: number;
    keyPrefix: string;
};
declare class CollabServer {
    readonly options: ServerOptions & typeof defualtOptions;
    private readonly streams;
    private readonly dirty;
    private constructor();
    /** Construct a new server. This will connect to Redis. */
    static of(options: ServerOptions): Promise<CollabServer>;
    /** Handle an incoming collaboration message. */
    handle(id: string, message: Message): Promise<any>;
    /** Read the latest version of a document from the server. */
    getDocument(id: string): Promise<Checkpoint>;
    /** Periodically checkpoint old documents by reading the dirty list. */
    compactionTask(): Promise<void>;
    private getDocumentInternal;
    private pullUpdates;
    private pushUpdates;
    private runCompaction;
}

export { Checkpoint, CollabServer, Message, ServerOptions };
